/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package org.milaifontanals.persistence;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Properties;
import org.basex.api.client.ClientSession;

/**
 * Capa de persistència per la gestió de l'empresa amb els seus departaments i
 * empleats, segons el model definit en el projecte Empresa.
 *
 * L'aplicació que usi la capa de persistència ha de disposar d'un fitxer de
 * propietats (veure documentació dels constructors) accessible via load(Reader)
 * amb les propietats següents:<BR>
 * host : Màquina (nom / ip) on és el SGBD BaseX<BR>
 * port : Port pel que escolta el SGBD BaseX<BR>
 * user : Usuari<BR>
 * password: Contrasenya<BR>
 *
 * @author Isidre Guixà
 */
public class EPBaseX {

    private ClientSession con;

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats de nom EPBaseX.properties.
     *
     * @throws EPBaseXException si hi ha algun problema en el fitxer de
     * configuració
     */
    public EPBaseX() {
        this("EPBaseX.properties");
    }

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats, i en cas de ser null o buit cercarà
     * el fitxer de nom EPBaseX.properties.
     *
     * @param nomFitxerPropietats
     * @throws EPBaseXException si hi ha algun problema en el fitxer de
     * configuració o si no s'estableix la connexió
     */
    public EPBaseX(String nomFitxerPropietats) {
        if (nomFitxerPropietats == null || nomFitxerPropietats.equals("")) {
            nomFitxerPropietats = "EPBaseX.properties";
        }
        Properties props = new Properties();
        try {
            props.load(new FileReader(nomFitxerPropietats));
        } catch (FileNotFoundException ex) {
            throw new EPBaseXException("No es troba fitxer de propietats " + nomFitxerPropietats, ex);
        } catch (IOException ex) {
            throw new EPBaseXException("Error en carregar fitxer de propietats " + nomFitxerPropietats, ex);
        }
        String host = props.getProperty("host");
        int port = 0;
        try {
            port = Integer.parseInt(props.getProperty("port"));
        } catch (NumberFormatException nfe) {
            throw new EPBaseXException("Port obligatori en fitxer de propietats i amb valor enter vàlid");
        }
        String user = props.getProperty("user");
        String password = props.getProperty("password");

        try {
            con = new ClientSession(host, port, user, password);
        } catch (IOException ex) {
            throw new EPBaseXException("No s'ha pogut establir la connexió", ex);
        }
    }
    
    /**
     * Tanca la capa de persistència, tancant la connexió amb la BD.
     *
     * @throws EPBaseXException si hi ha algun problema en tancar la connexió
     */
    public void closeCapa() {
        if (con!=null) {
            try {
                con.close();
            } catch (IOException ex) {
                throw new EPBaseXException("Error en tancar la connexió",ex);
            } finally {
                con = null;
            }
        }
    }
}
