/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package org.milaifontanals.persistence;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.basex.api.client.ClientQuery;
import org.basex.api.client.ClientSession;
import org.milaifontanals.empresa.Empresa;

/**
 * Capa de persistència per la gestió de l'empresa amb els seus departaments i
 * empleats, segons el model definit en el projecte Empresa.
 *
 * L'aplicació que usi la capa de persistència ha de disposar d'un fitxer de
 * propietats (veure documentació dels constructors) accessible via load(Reader)
 * amb les propietats següents:<BR>
 * host : Màquina (nom / ip) on és el SGBD BaseX<BR>
 * port : Port pel que escolta el SGBD BaseX<BR>
 * user : Usuari<BR>
 * password: Contrasenya<BR>
 * path: Ruta del fitxer XML en el que es basa la capa de persistència
 * incorporant el nom de la BD en format doc("BD/cami/nomFitxerXml"). Les dades
 * es troben en un unic fitxer que segueix empresa.dtd
 *
 * @author Isidre Guixà
 */
public class EPBaseX {

    private ClientSession con;
    private String path;

    private static SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats de nom EPBaseX.properties.
     *
     * @throws EPBaseXException si hi ha algun problema en el fitxer de
     * configuració
     */
    public EPBaseX() {
        this("EPBaseX.properties");
    }

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats, i en cas de ser null o buit cercarà
     * el fitxer de nom EPBaseX.properties.
     *
     * @param nomFitxerPropietats
     * @throws EPBaseXException si hi ha algun problema en el fitxer de
     * configuració o si no s'estableix la connexió
     */
    public EPBaseX(String nomFitxerPropietats) {
        if (nomFitxerPropietats == null || nomFitxerPropietats.equals("")) {
            nomFitxerPropietats = "EPBaseX.properties";
        }
        Properties props = new Properties();
        try {
            props.load(new FileReader(nomFitxerPropietats));
        } catch (FileNotFoundException ex) {
            throw new EPBaseXException("No es troba fitxer de propietats " + nomFitxerPropietats, ex);
        } catch (IOException ex) {
            throw new EPBaseXException("Error en carregar fitxer de propietats " + nomFitxerPropietats, ex);
        }
        String host = props.getProperty("host");
        int port = 0;
        try {
            port = Integer.parseInt(props.getProperty("port"));
        } catch (NumberFormatException nfe) {
            throw new EPBaseXException("Port obligatori en fitxer de propietats i amb valor enter vàlid");
        }
        String user = props.getProperty("user");
        String password = props.getProperty("password");

        try {
            con = new ClientSession(host, port, user, password);
        } catch (IOException ex) {
            throw new EPBaseXException("No s'ha pogut establir la connexió", ex);
        }

        path = props.getProperty("path");
        if (path == null || path.equals("")) {
            closeCapa();
            throw new EPBaseXException("El fitxer de propietats no incorpora la propietat path amb valor adequat");
        }

        ClientQuery q = null;

        try {
            q = con.query(path + "/x");
            q.execute();
        } catch (IOException ex) {
            closeCapa();
            throw new EPBaseXException("Error en comprovar existencia de " + path + " en el SGBD.", ex);
        } finally {
            if (con != null) {
                tancarQuery(q);
            }
        }
    }

    /**
     * Tanca la capa de persistència, tancant la connexió amb la BD.
     *
     * @throws EPBaseXException si hi ha algun problema en tancar la connexió
     */
    public void closeCapa() {
        if (con != null) {
            try {
                con.close();
            } catch (IOException ex) {
                throw new EPBaseXException("Error en tancar la connexió", ex);
            } finally {
                con = null;
            }
        }
    }

    public Empresa getEmpresa() {
        String nom = null;
        Calendar dataCreacio = null;
        ClientQuery q = null;
        try {
            q = con.query(path + "/empresa/nom/string()");
            nom = q.execute();
            q = con.query(path + "/empresa/dataCreacio/string()");
            String aux = q.execute();
            dataCreacio = new GregorianCalendar();
            dataCreacio.setTime(sdf.parse(aux));
        } catch (IOException | ParseException ex) {
            throw new EPBaseXException("Error en getEmpresa", ex);
        } finally {
            tancarQuery(q);
        }
        return new Empresa(nom, dataCreacio);
    }

    /**
     *
     * MÈTODES PRIVATS, d'ajut pel desenvolupament de la capa
     *
     */
    /**
     * Tanca la query indicada per paràmetre.
     *
     * @param q Query a tancar
     */
    private void tancarQuery(ClientQuery q) {
        if (q != null) {
            try {
                q.close();
            } catch (IOException ex) {
                throw new EPBaseXException("Error en close query", ex);
            }
        }
    }

}
