/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package org.milaifontanals.persistence;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.milaifontanals.empresa.Departament;
import org.milaifontanals.empresa.Empleat;
import org.milaifontanals.empresa.Empresa;
import ru.ispras.sedna.driver.DatabaseManager;
import ru.ispras.sedna.driver.DriverException;
import ru.ispras.sedna.driver.SednaConnection;
import ru.ispras.sedna.driver.SednaStatement;

/**
 * Capa de persistència per la gestió de l'empresa amb els seus departaments i
 * empleats, segons el model definit en el projecte Empresa.
 *
 * L'aplicació que usi la capa de persistència ha de disposar d'un fitxer de
 * propietats (veure documentació dels constructors) accessible via load(Reader)
 * amb les propietats següents:<BR>
 * host : Màquina (nom / ip) on és el SGBD Sedna<BR>
 * port : Port pel que escolta el SGBD Sedna<BR>
 * bd : Nom de la base de dades<BR>
 * user : Usuari<BR>
 * password: Contrasenya<BR>
 * path: Ruta del fitxer XML en el que es basa la capa de persistència
 * incorporant el nom de la BD en format doc("nomFitxerXml","nomCol·lecció) o
 * doc("nomFitxerXml") si no és dins cap col·lecció
 *
 * @author Isidre Guixà
 */
public class EPSedna {

    private SednaConnection con;
    private String path;

    /* Variable per control de transaccions */
    private boolean transOn;

    /* Variable per formatar les dates segons format requerit */
    private static SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");

    /* Variables per conèixer els objectes que tenim en memòria */
    private Empresa empresa;
    private HashMap<Integer, Departament> hmDept = new HashMap();
    private HashMap<Integer, Empleat> hmEmp = new HashMap();

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats de nom EPSedna.properties.
     *
     * @throws EPSednaException si hi ha algun problema en el fitxer de
     * configuració o en establir connexió
     */
    public EPSedna() {
        this("EPSedna.properties");
    }

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats, i en cas de ser null o buit cercarà
     * el fitxer de nom EPSedna.properties.
     *
     * @param nomFitxerPropietats
     * @throws EPSednaException si hi ha algun problema en el fitxer de
     * configuració o si no s'estableix la connexió
     */
    public EPSedna(String nomFitxerPropietats) {
        if (nomFitxerPropietats == null || nomFitxerPropietats.equals("")) {
            nomFitxerPropietats = "EPSedna.properties";
        }
        Properties props = new Properties();
        try {
            props.load(new FileReader(nomFitxerPropietats));
        } catch (FileNotFoundException ex) {
            throw new EPSednaException("No es troba fitxer de propietats " + nomFitxerPropietats, ex);
        } catch (IOException ex) {
            throw new EPSednaException("Error en carregar fitxer de propietats " + nomFitxerPropietats, ex);
        }
        String host = props.getProperty("host");
        String port = props.getProperty("port");
        String user = props.getProperty("user");
        String password = props.getProperty("password");
        String database = props.getProperty("database");

        try {
            String url = host + ":" + port;
            con = DatabaseManager.getConnection(url, database, user, password);
        } catch (DriverException ex) {
            throw new EPSednaException("No s'ha pogut establir la connexió", ex);
        }

        path = props.getProperty("path");
        if (path == null || path.equals("")) {
            closeCapa();
            throw new EPSednaException("El fitxer de propietats no incorpora la propietat path amb valor adequat");
        }

        SednaStatement q = null;

        try {
            q = con.createStatement();
            con.begin();
            transOn = true;
            q.execute(path + "/x");
        } catch (DriverException ex) {
            closeCapa();
            throw new EPSednaException("Error en comprovar existencia de " + path + " en el SGBD.", ex);
        }
    }

    /**
     * Tanca la capa de persistència, tancant la connexió amb la BD.
     *
     * @throws EPSEdnaException si hi ha algun problema en tancar la connexió
     */
    public void closeCapa() {
        if (con != null) {
            if (transOn) {
                try {
                    con.rollback();
                    transOn = false;
                } catch (DriverException ex) {
                }
            }
            try {
                con.close();
            } catch (DriverException ex) {
                throw new EPSednaException("Error en tancar la connexió", ex);
            } finally {
                con = null;
            }
        }
    }

}
