/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package org.milaifontanals.persistence;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Properties;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import org.milaifontanals.empresa.Departament;
import org.milaifontanals.empresa.Empleat;
import org.milaifontanals.empresa.Empresa;
import ru.ispras.sedna.driver.DatabaseManager;
import ru.ispras.sedna.driver.DriverException;
import ru.ispras.sedna.driver.SednaConnection;
import ru.ispras.sedna.driver.SednaStatement;

/**
 * Capa de persistència per la gestió de l'empresa amb els seus departaments i
 * empleats, segons el model definit en el projecte Empresa.
 *
 * L'aplicació que usi la capa de persistència ha de disposar d'un fitxer de
 * propietats (veure documentació dels constructors) accessible via load(Reader)
 * amb les propietats següents:<BR>
 * host : Màquina (nom / ip) on és el SGBD Sedna<BR>
 * port : Port pel que escolta el SGBD Sedna<BR>
 * bd : Nom de la base de dades<BR>
 * user : Usuari<BR>
 * password: Contrasenya<BR>
 * path: Ruta del fitxer XML en el que es basa la capa de persistència
 * incorporant el nom de la BD en format doc("nomFitxerXml","nomCol·lecció) o
 * doc("nomFitxerXml") si no és dins cap col·lecció
 *
 * @author Isidre Guixà
 */
public class EPSedna {

    private SednaConnection con;
    private String path;

    /* Variable per control de transaccions */
    private boolean transOn;

    /* Variable per formatar les dates segons format requerit */
    private static SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");

    /* Variables per conèixer els objectes que tenim en memòria */
    private Empresa empresa;
    private HashMap<Integer, Departament> hmDept = new HashMap();
    private HashMap<Integer, Empleat> hmEmp = new HashMap();

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats de nom EPSedna.properties.
     *
     * @throws EPSednaException si hi ha algun problema en el fitxer de
     * configuració o en establir connexió
     */
    public EPSedna() {
        this("EPSedna.properties");
    }

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats, i en cas de ser null o buit cercarà
     * el fitxer de nom EPSedna.properties.
     *
     * @param nomFitxerPropietats
     * @throws EPSednaException si hi ha algun problema en el fitxer de
     * configuració o si no s'estableix la connexió
     */
    public EPSedna(String nomFitxerPropietats) {
        if (nomFitxerPropietats == null || nomFitxerPropietats.equals("")) {
            nomFitxerPropietats = "EPSedna.properties";
        }
        Properties props = new Properties();
        try {
            props.load(new FileReader(nomFitxerPropietats));
        } catch (FileNotFoundException ex) {
            throw new EPSednaException("No es troba fitxer de propietats " + nomFitxerPropietats, ex);
        } catch (IOException ex) {
            throw new EPSednaException("Error en carregar fitxer de propietats " + nomFitxerPropietats, ex);
        }
        String host = props.getProperty("host");
        String port = props.getProperty("port");
        String user = props.getProperty("user");
        String password = props.getProperty("password");
        String database = props.getProperty("database");

        try {
            String url = host + ":" + port;
            con = DatabaseManager.getConnection(url, database, user, password);
        } catch (DriverException ex) {
            throw new EPSednaException("No s'ha pogut establir la connexió", ex);
        }

        path = props.getProperty("path");
        if (path == null || path.equals("")) {
            closeCapa();
            throw new EPSednaException("El fitxer de propietats no incorpora la propietat path amb valor adequat");
        }

        SednaStatement q = null;

        try {
            q = con.createStatement();
            con.begin();
            transOn = true;
            q.execute(path + "/x");
        } catch (DriverException ex) {
            closeCapa();
            throw new EPSednaException("Error en comprovar existencia de " + path + " en el SGBD.", ex);
        }
    }

    /**
     * Tanca la capa de persistència, tancant la connexió amb la BD.
     *
     * @throws EPSEdnaException si hi ha algun problema en tancar la connexió
     */
    public void closeCapa() {
        if (con != null) {
            if (transOn) {
                try {
                    con.rollback();
                    transOn = false;
                } catch (DriverException ex) {
                }
            }
            try {
                con.close();
            } catch (DriverException ex) {
                throw new EPSednaException("Error en tancar la connexió", ex);
            } finally {
                con = null;
            }
        }
    }

    /**
     * Tanca la transacció activa validant els canvis a la BD.
     *
     * @throws EPSednaException si hi ha algun problema
     */
    public void commit() {
        if (transOn) {
            try {
                con.commit();
                transOn = false;
            } catch (DriverException ex) {
                transOn = false;
                throw new EPSednaException("Error en fer commit.", ex);
            }
        }
    }

    /**
     * Tanca la transacció activa sense validar els canvis a la BD.
     *
     * @throws EPSednaException si hi ha algun problema
     */
    public void rollback() {
        if (transOn) {
            try {
                con.rollback();
                transOn = false;
            } catch (DriverException ex) {
                transOn = false;
                throw new EPSednaException("Error en fer rollback.", ex);
            }
        }
    }

    /**
     * Recupera l'objecte Empresa
     *
     * @return L'objecte Empresa
     * @throws EPSednaException si hi ha algun problema en recuperar l'empresa
     */
    public Empresa getEmpresa() {
        if (empresa != null) {
            return empresa;
        }
        // Si encara no teniem l'empresa en memòria:
        SednaStatement q = null;
        String nom = null;
        Calendar dataCreacio = null;
        try {
            obrirTrans();
            q = con.createStatement();
            q.execute(path + "/empresa/nom/string()");
            nom = q.getSerializedResult().next();
            q.execute(path + "/empresa/dataCreacio/string()");
            String aux = q.getSerializedResult().next();
            dataCreacio = new GregorianCalendar();
            dataCreacio.setTime(sdf.parse(aux));
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Error en getEmpresa", ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en getEmpresa", ex);
        }
        empresa = new Empresa(nom, dataCreacio);
        return empresa;
    }

    /**
     * Retorna el departament de codi indicat, si existix o null en cas
     * contrari.
     *
     * @param codi Codi del departament a obtenir
     * @return El departament o null
     * @throws EPSednaException si es produeix algun error
     */
    public Departament getDepartament(int codi) {
        if (codi <= 0) {
            throw new EPSednaException("Intent de recuperar departament amb codi " + codi + " erroni");
        }
        Departament d = hmDept.get(codi);
        if (d != null) {
            return d;
        }
        // Si encara no teniem el departament en memòria:
        SednaStatement q = null;
        try {
            obrirTrans();
            String cad = path + "//departaments/dept[@codi='d" + codi + "']";
//            System.out.println(cad);
            q = con.createStatement();
            q.execute(cad);
            String resultat = q.getSerializedResult().next();
            if (resultat == null) {
                return null;
            }
            // Dins "resultat" hi ha un node <dept>
            SAXBuilder sb = new SAXBuilder();
            Document doc = sb.build(new StringReader(resultat));
            Element xDept = doc.getRootElement();
            String nom = xDept.getChildText("nom");
            String loc = xDept.getChildText("localitat");
            d = new Departament(codi, nom, loc);
            hmDept.put(codi, d);
            return d;
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        }
    }

    /**
     * Intenta recuperar l'empleat amb codi indicat per paràmetre.
     *
     * @param codi de l'empleat a recuperar
     * @return L'objecte Empleat trobat o null si no existeix
     * @throws EPSednaException si es produeix algun error (introduir un codi
     * inadequat genera també aquesta excepció)
     */
    public Empleat getEmpleat(int codi) {
        if (codi <= 0) {
            throw new EPSednaException("Intent de cercar un empleat amb codi " + codi + " inadequat");
        }
        Empleat e = hmEmp.get(codi);
        if (e != null) {
            return e;
        }
        // Si encara no teniem l'empleat en memòria:
        SednaStatement q = null;
        try {
            obrirTrans();
            String cad = path + "//empleats/emp[@codi='e" + codi + "']";
//            System.out.println(cad);
            q = con.createStatement();
            q.execute(cad);
            String resultat = q.getSerializedResult().next();
            if (resultat == null) {
                return null;
            }
            // Dins "resultat" hi ha un node <emp>
            SAXBuilder sb = new SAXBuilder();
            Document doc = sb.build(new StringReader(resultat));
            Element xEmp = doc.getRootElement();
            String cognom = xEmp.getChildText("cognom");
            String ofici = xEmp.getChildText("ofici");
            String sDataAlta = xEmp.getChildText("dataAlta");
            Calendar auxDataAlta = null;
            if (sDataAlta != null && !sDataAlta.equals("")) {
                auxDataAlta = new GregorianCalendar();
                try {
                    auxDataAlta.setTime(sdf.parse(sDataAlta));
                } catch (ParseException ex) {
                }
            }
            String sSalari = xEmp.getChildText("salari");
            Double auxSalari = null;
            if (sSalari != null) {
                auxSalari = Double.parseDouble(sSalari);
            }
            String sComissio = xEmp.getChildText("comissio");
            Double auxComissio = null;
            if (sComissio != null) {
                auxComissio = Double.parseDouble(sComissio);
            }
            String sDept = xEmp.getAttributeValue("dept");
            Departament auxDept = null;
            if (sDept != null) {
                auxDept = getDepartament(Integer.parseInt(sDept.substring(1)));
            }
            String sCap = xEmp.getAttributeValue("cap");
            Empleat auxCap = null;
            if (sCap != null) {
                auxCap = getEmpleat(Integer.parseInt(sCap.substring(1)));
            }
            e = new Empleat(codi, cognom, ofici, auxDataAlta, auxSalari,
                    auxComissio, auxCap, auxDept);
            hmEmp.put(codi, e);
            return e;
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        }
    }

    /**
     * Informa de l'existència d'empleat amb el codi indicat per paràmetre.
     *
     * @return Valor booleà que indica l'existència de l'empleat indicat
     * @throws EPSednaException si es produeix algun error (introduir un codi
     * inadequat genera també aquesta excepció)
     */
    public boolean existeixEmpleat(int codi) {
        if (codi <= 0) {
            throw new EPSednaException("Intent de cercar un empleat amb codi " + codi + " inadequat");
        }
        if (hmEmp.containsKey(codi)) {
            return true;
        }
        SednaStatement q = null;
        try {
            obrirTrans();
            String cad = path + "//emp[@codi='e" + codi + "']/@codi";
            q = con.createStatement();
            q.execute(cad);
            // Cal analitzar el resultat obtingut
            return q.getSerializedResult().next() != null;
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Problema en comprovar existència d'empleat amb codi " + codi, ex);
        }
    }


    /**
     *
     * MÈTODES PRIVATS, d'ajut pel desenvolupament de la capa
     *
     */
    private void obrirTrans() throws DriverException {
        if (!transOn) {
            con.begin();
            transOn = true;
        }
    }

}
