/*
 * Click nbfs://nbhost/SystemFileSystem/Templates/Licenses/license-default.txt to change this license
 * Click nbfs://nbhost/SystemFileSystem/Templates/Classes/Class.java to edit this template
 */
package org.milaifontanals.persistence;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import org.milaifontanals.empresa.Departament;
import org.milaifontanals.empresa.Empleat;
import org.milaifontanals.empresa.Empresa;
import ru.ispras.sedna.driver.DatabaseManager;
import ru.ispras.sedna.driver.DriverException;
import ru.ispras.sedna.driver.SednaConnection;
import ru.ispras.sedna.driver.SednaSerializedResult;
import ru.ispras.sedna.driver.SednaStatement;

/**
 * Capa de persistència per la gestió de l'empresa amb els seus departaments i
 * empleats, segons el model definit en el projecte Empresa.
 *
 * L'aplicació que usi la capa de persistència ha de disposar d'un fitxer de
 * propietats (veure documentació dels constructors) accessible via load(Reader)
 * amb les propietats següents:<BR>
 * host : Màquina (nom / ip) on és el SGBD Sedna<BR>
 * port : Port pel que escolta el SGBD Sedna<BR>
 * bd : Nom de la base de dades<BR>
 * user : Usuari<BR>
 * password: Contrasenya<BR>
 * path: Ruta del fitxer XML en el que es basa la capa de persistència
 * incorporant el nom de la BD en format doc("nomFitxerXml","nomCol·lecció) o
 * doc("nomFitxerXml") si no és dins cap col·lecció
 *
 * @author Isidre Guixà
 */
public class EPSedna {

    private SednaConnection con;
    private String path;

    /* Variable per control de transaccions */
    private boolean transOn;

    /* Variable per formatar les dates segons format requerit */
    private static SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");

    /* Variables per conèixer els objectes que tenim en memòria */
    private Empresa empresa;
    private HashMap<Integer, Departament> hmDept = new HashMap();
    private HashMap<Integer, Empleat> hmEmp = new HashMap();

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats de nom EPSedna.properties.
     *
     * @throws EPSednaException si hi ha algun problema en el fitxer de
     * configuració o en establir connexió
     */
    public EPSedna() {
        this("EPSedna.properties");
    }

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats, i en cas de ser null o buit cercarà
     * el fitxer de nom EPSedna.properties.
     *
     * @param nomFitxerPropietats
     * @throws EPSednaException si hi ha algun problema en el fitxer de
     * configuració o si no s'estableix la connexió
     */
    public EPSedna(String nomFitxerPropietats) {
        if (nomFitxerPropietats == null || nomFitxerPropietats.equals("")) {
            nomFitxerPropietats = "EPSedna.properties";
        }
        Properties props = new Properties();
        try {
            props.load(new FileReader(nomFitxerPropietats));
        } catch (FileNotFoundException ex) {
            throw new EPSednaException("No es troba fitxer de propietats " + nomFitxerPropietats, ex);
        } catch (IOException ex) {
            throw new EPSednaException("Error en carregar fitxer de propietats " + nomFitxerPropietats, ex);
        }
        String host = props.getProperty("host");
        String port = props.getProperty("port");
        String user = props.getProperty("user");
        String password = props.getProperty("password");
        String database = props.getProperty("database");

        try {
            String url = host + ":" + port;
            con = DatabaseManager.getConnection(url, database, user, password);
        } catch (DriverException ex) {
            throw new EPSednaException("No s'ha pogut establir la connexió", ex);
        }

        path = props.getProperty("path");
        if (path == null || path.equals("")) {
            closeCapa();
            throw new EPSednaException("El fitxer de propietats no incorpora la propietat path amb valor adequat");
        }

        SednaStatement q = null;

        try {
            q = con.createStatement();
            con.begin();
            transOn = true;
            q.execute(path + "/x");
        } catch (DriverException ex) {
            closeCapa();
            throw new EPSednaException("Error en comprovar existencia de " + path + " en el SGBD.", ex);
        }
    }

    /**
     * Tanca la capa de persistència, tancant la connexió amb la BD.
     *
     * @throws EPSEdnaException si hi ha algun problema en tancar la connexió
     */
    public void closeCapa() {
        if (con != null) {
            if (transOn) {
                try {
                    con.rollback();
                    transOn = false;
                } catch (DriverException ex) {
                }
            }
            try {
                con.close();
            } catch (DriverException ex) {
                throw new EPSednaException("Error en tancar la connexió", ex);
            } finally {
                con = null;
            }
        }
    }

    /**
     * Tanca la transacció activa validant els canvis a la BD.
     *
     * @throws EPSednaException si hi ha algun problema
     */
    public void commit() {
        if (transOn) {
            try {
                con.commit();
                transOn = false;
            } catch (DriverException ex) {
                transOn = false;
                throw new EPSednaException("Error en fer commit.", ex);
            }
        }
    }

    /**
     * Tanca la transacció activa sense validar els canvis a la BD.
     *
     * @throws EPSednaException si hi ha algun problema
     */
    public void rollback() {
        if (transOn) {
            try {
                con.rollback();
                transOn = false;
            } catch (DriverException ex) {
                transOn = false;
                throw new EPSednaException("Error en fer rollback.", ex);
            }
        }
    }

    /**
     * Recupera l'objecte Empresa
     *
     * @return L'objecte Empresa
     * @throws EPSednaException si hi ha algun problema en recuperar l'empresa
     */
    public Empresa getEmpresa() {
        if (empresa != null) {
            return empresa;
        }
        // Si encara no teniem l'empresa en memòria:
        SednaStatement q = null;
        String nom = null;
        Calendar dataCreacio = null;
        try {
            obrirTrans();
            q = con.createStatement();
            q.execute(path + "/empresa/nom/string()");
            nom = q.getSerializedResult().next();
            q.execute(path + "/empresa/dataCreacio/string()");
            String aux = q.getSerializedResult().next();
            dataCreacio = new GregorianCalendar();
            dataCreacio.setTime(sdf.parse(aux));
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Error en getEmpresa", ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en getEmpresa", ex);
        }
        empresa = new Empresa(nom, dataCreacio);
        return empresa;
    }

    /**
     * Retorna el departament de codi indicat, si existix o null en cas
     * contrari.
     *
     * @param codi Codi del departament a obtenir
     * @return El departament o null
     * @throws EPSednaException si es produeix algun error
     */
    public Departament getDepartament(int codi) {
        if (codi <= 0) {
            throw new EPSednaException("Intent de recuperar departament amb codi " + codi + " erroni");
        }
        Departament d = hmDept.get(codi);
        if (d != null) {
            return d;
        }
        // Si encara no teniem el departament en memòria:
        SednaStatement q = null;
        try {
            obrirTrans();
            String cad = path + "//departaments/dept[@codi='d" + codi + "']";
//            System.out.println(cad);
            q = con.createStatement();
            q.execute(cad);
            String resultat = q.getSerializedResult().next();
            if (resultat == null) {
                return null;
            }
            // Dins "resultat" hi ha un node <dept>
            SAXBuilder sb = new SAXBuilder();
            Document doc = sb.build(new StringReader(resultat));
            Element xDept = doc.getRootElement();
            String nom = xDept.getChildText("nom");
            String loc = xDept.getChildText("localitat");
            d = new Departament(codi, nom, loc);
            hmDept.put(codi, d);
            return d;
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        }
    }

    /**
     * Intenta recuperar l'empleat amb codi indicat per paràmetre.
     *
     * @param codi de l'empleat a recuperar
     * @return L'objecte Empleat trobat o null si no existeix
     * @throws EPSednaException si es produeix algun error (introduir un codi
     * inadequat genera també aquesta excepció)
     */
    public Empleat getEmpleat(int codi) {
        if (codi <= 0) {
            throw new EPSednaException("Intent de cercar un empleat amb codi " + codi + " inadequat");
        }
        Empleat e = hmEmp.get(codi);
        if (e != null) {
            return e;
        }
        // Si encara no teniem l'empleat en memòria:
        SednaStatement q = null;
        try {
            obrirTrans();
            String cad = path + "//empleats/emp[@codi='e" + codi + "']";
//            System.out.println(cad);
            q = con.createStatement();
            q.execute(cad);
            String resultat = q.getSerializedResult().next();
            if (resultat == null) {
                return null;
            }
            // Dins "resultat" hi ha un node <emp>
            SAXBuilder sb = new SAXBuilder();
            Document doc = sb.build(new StringReader(resultat));
            Element xEmp = doc.getRootElement();
            String cognom = xEmp.getChildText("cognom");
            String ofici = xEmp.getChildText("ofici");
            String sDataAlta = xEmp.getChildText("dataAlta");
            Calendar auxDataAlta = null;
            if (sDataAlta != null && !sDataAlta.equals("")) {
                auxDataAlta = new GregorianCalendar();
                try {
                    auxDataAlta.setTime(sdf.parse(sDataAlta));
                } catch (ParseException ex) {
                }
            }
            String sSalari = xEmp.getChildText("salari");
            Double auxSalari = null;
            if (sSalari != null) {
                auxSalari = Double.parseDouble(sSalari);
            }
            String sComissio = xEmp.getChildText("comissio");
            Double auxComissio = null;
            if (sComissio != null) {
                auxComissio = Double.parseDouble(sComissio);
            }
            String sDept = xEmp.getAttributeValue("dept");
            Departament auxDept = null;
            if (sDept != null) {
                auxDept = getDepartament(Integer.parseInt(sDept.substring(1)));
            }
            String sCap = xEmp.getAttributeValue("cap");
            Empleat auxCap = null;
            if (sCap != null) {
                auxCap = getEmpleat(Integer.parseInt(sCap.substring(1)));
            }
            e = new Empleat(codi, cognom, ofici, auxDataAlta, auxSalari,
                    auxComissio, auxCap, auxDept);
            hmEmp.put(codi, e);
            return e;
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        }
    }

    /**
     * Informa de l'existència d'empleat amb el codi indicat per paràmetre.
     *
     * @return Valor booleà que indica l'existència de l'empleat indicat
     * @throws EPSednaException si es produeix algun error (introduir un codi
     * inadequat genera també aquesta excepció)
     */
    public boolean existeixEmpleat(int codi) {
        if (codi <= 0) {
            throw new EPSednaException("Intent de cercar un empleat amb codi " + codi + " inadequat");
        }
        if (hmEmp.containsKey(codi)) {
            return true;
        }
        SednaStatement q = null;
        try {
            obrirTrans();
            String cad = path + "//emp[@codi='e" + codi + "']/@codi";
            q = con.createStatement();
            q.execute(cad);
            // Cal analitzar el resultat obtingut
            return q.getSerializedResult().next() != null;
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Problema en comprovar existència d'empleat amb codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en intentar recuperar departament de codi " + codi, ex);
        }
    }

    /**
     * Facilita el nombre de subordinats d'un empleat
     *
     * @param codi de l'empleat al què comptabilitzar els subordinats
     * @return El nombre de subordinats de l'empleat
     * @throws EPSednaException si es produeix algun error (introduir un codi
     * inadequat o la inexistència d'empleat amb el codi indicat genera també
     * aquesta excepció)
     */
    public int comptarSubordinats(int codi) {
        if (!existeixEmpleat(codi)) {
            throw new EPSednaException("Intent de comptar subordinats d'empleat amb codi " + codi + " inexistent");
        }
        // Cal comptar:
        String cad = "count(" + path + "//emp[@cap='e" + codi + "'])";
        SednaStatement q = null;
        try {
            obrirTrans();
            q = con.createStatement();
            q.execute(cad);
            return Integer.parseInt(q.getSerializedResult().next());
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Problema en comprovar existència d'empleat amb codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en comptar els subordinats d'empleat de codi " + codi, ex);
        }
    }

    /**
     * Comprova si un empleat és subordinat directe o indirecte (varis nivells)
     * del cap
     *
     * @param cap
     * @param emp
     * @return true si emp és subordinat (directe o indirecte) de cap i false en
     * cas contrari
     */
    public boolean esSubordinatDirecteIndirecte(int cap, int emp) {
        if (!this.existeixEmpleat(cap)) {
            throw new EPSednaException("No existeix empleat amb codi " + cap);
        }
        if (!this.existeixEmpleat(emp)) {
            throw new EPSednaException("No existeix empleat amb codi " + emp);
        }
        /* Instrucció que facilita tots els subordinats directes del "cap" */
        String cad = path + "//emp[@cap='e" + cap + "']/@codi/string()";
        SednaStatement q = null;
        try {
            obrirTrans();
            q = con.createStatement();
            q.execute(cad);
            SednaSerializedResult s = q.getSerializedResult();
            String eCodi = s.next();
            if (eCodi == null) {
                return false;
            }
            List<String> subordinats = new ArrayList();
            while (eCodi != null) {
                subordinats.add(eCodi);
                eCodi = s.next();
            }
            /* 
            ALERTA 1:
                Mentre s’està processant un SednaSerializedResult, NO es pot processar 
                un altre SednaSerializedResult. En cas de fer-ho, el primer queda en un 
                estat inconsistent, i en intentar continuar el seu procés, el programa 
                queda “penjat” (no respon). Per això hem hagut de recollir primer tots
                els codis en una llista, doncs si agafem només un codi i el processem i
                en fer la crida recursiva es torna a processar un SednaSerializedResult,
                el SednaSerializedResult anterior quedaria inconsistent.
            ALERTA 2:
                En un SednaSerializedResult, a partir del 2n resultat comencen amb
                un salt de línia, que hem d'eliminar i que és diferent segons S.O.
             */

            for (String eCodiBis : subordinats) {
                eCodiBis = eCodiBis.substring(eCodiBis.indexOf("e") + 1);
                //La instrucció anterior elimina tot el què hi hagi fins la "e" inclosa
                int codi = Integer.parseInt(eCodiBis);
                if (codi == emp) {
                    return true;
                } else {
                    boolean aux = esSubordinatDirecteIndirecte(codi, emp);
                    if (aux) {
                        return aux;
                    }
                }
            }
            return false;
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Error en cercar els subordinats d'empleat de codi " + cap, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Error en cercar els subordinats d'empleat de codi " + cap, ex);
        }
    }

    /**
     * Intenta eliminar un empleat de codi indicat per paràmetre.
     *
     * @param codi de l'empleat a eliminar
     * @param actCap: Indica l'actuació a tenir en compte vers els empleats dels
     * que pugui ser el cap. No té cap efecte si l'empleat no és cap de ningú.
     * Actuacions:<BR>
     * Valor 0: Deixar els empleats que el tenen per cap, "SENSE" cap<BR>
     * Valor &lt;0: Generar excepció indicant que no ha estat possible
     * l'eliminació per ser cap d'altres empleats<BR>
     * Valor &gt;0: Canviar el cap dels empleats afectats per l'empleat amb codi
     * = al valor del paràmetre actCap (que ha d'existir)<BR>
     *
     */
    public void eliminarEmpleat(int codi, int actCap) {
        if (!existeixEmpleat(codi)) {
            throw new EPSednaException("Intent d'eliminar empleat amb codi " + codi + " inexistent");
        }
        if (codi == actCap) {
            throw new EPSednaException("Intent d'eliminar empleat amb codi " + codi
                    + " reassignant els seus subordinats a ell mateix ");
        }
        if (actCap < 0) {
            throw new EPSednaException("S'ha invocat mètode eliminarEmpleat amb 2n paràmetre erroni");
        }
        // Sabem que l'empleat existeix
        List<String> llCad = new ArrayList();
        if (actCap == 0) {
            /* Deixar els subordinats sense cap i eliminar l'empleat */
            // Instrucció per eliminar dels subordinats, l'atribut "cap": 
            llCad.add("update delete " + path + "//emp[@cap='e" + codi + "']/@cap");
            // Instrucció per eliminar el propi empleat
            llCad.add("update delete " + path + "//emp[@codi='e" + codi + "']");
            // cad conté la seqûència d'instruccions per deixar els subordinats sense cap
            // i eliminar l'empleat
        } else {
            /* Hem d'intentar substituir el cap dels subordinats (pot ser que no n'hi hagi) */
            if (!this.existeixEmpleat(actCap)) {
                throw new EPSednaException("Intent d'eliminar empleat amb codi " + codi
                        + " reassignant els seus subordinats a empleat de codi " + actCap + " inexistent");
            }
            if (this.esSubordinatDirecteIndirecte(codi, actCap)) {
                throw new EPSednaException("Intent d'eliminar empleat amb codi " + codi
                        + " reassignant els seus subordinats a empleat de codi " + actCap
                        + " que és un subordinat (directe o indirecte) del cap");
            }
            /* Cal canviar el cap als subordinats directes i eliminar l'empleat 
               En Sedna no hi ha possibilitat de substituir el valor d'un atribut
               i cal substituir l'atribut sencer amb "update replace"
             */
            llCad.add("update replace $n in " + path + "//emp[@cap='e" + codi + "']/@cap\n"
                    + "with (attribute cap {'e" + actCap + "'})");
            // Instrucció per eliminar el propi empleat
            llCad.add("update delete " + path + "//emp[@codi='e" + codi + "']");
            // cad conté la seqûència d'instruccions per canviar el cap als subordinats
            // i eliminar l'empleat
        }
        SednaStatement q = null;
        try {
            obrirTrans();
            q = con.createStatement();
            for (String cad : llCad) {
                q.execute(cad);
            }
            hmEmp.remove(codi);
            Collection empleats = hmEmp.values();
            Iterator<Empleat> iteEmp = null;
            if (actCap == 0) {
                // Cal eliminar el cap als possibles subordinats existents en memòria
                iteEmp = empleats.iterator();
                // ALERTA! Dins "else" següent, explicat per què NO es pot referenciar iteEmp després de crear "empleats"
                while (iteEmp.hasNext()) {
                    Empleat e = iteEmp.next();
                    if (e.getCap() != null && e.getCap().getCodi() == codi) {
                        e.setCap(null);
                    }
                }
            } else {
                // Cal actualitzar el cap als possibles subordinats existents en memòria
                // Primer cal tenir la referència que apunta al nou cap
                Empleat nouCap = getEmpleat(actCap);
                // En cas que actCap no hagués estat en memòria, el mètode getEmpleat el cerca i l'afegeix a iteEmp.
                // I Java NO permet recórrer un iterator d'un HashMap que hagi estat creat abans d'una modificació del HashMap
                // Es genera una ConcurrentModificationException!!!
                iteEmp = empleats.iterator();
                while (iteEmp.hasNext()) {
                    Empleat e = iteEmp.next();
                    if (e.getCap() != null && e.getCap().getCodi() == codi) {
                        e.setCap(nouCap);
                    }
                }
            }
        } catch (DriverException ex) {
            transOn = false;
            throw new EPSednaException("Problema en intentar eliminar empleat amb codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPSednaException("Problema en intentar eliminar empleat amb codi " + codi, ex);
        }
    }

    /**
     *
     * MÈTODES PRIVATS, d'ajut pel desenvolupament de la capa
     *
     */
    private void obrirTrans() throws DriverException {
        if (!transOn) {
            con.begin();
            transOn = true;
        }
    }

}
