package org.milaifontanals.persistence;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import javax.xml.namespace.QName;
import javax.xml.xquery.XQConnection;
import javax.xml.xquery.XQDataSource;
import javax.xml.xquery.XQException;
import javax.xml.xquery.XQExpression;
import javax.xml.xquery.XQItemType;
import javax.xml.xquery.XQPreparedExpression;
import javax.xml.xquery.XQResultSequence;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import org.milaifontanals.empresa.Departament;
import org.milaifontanals.empresa.Empleat;
import org.milaifontanals.empresa.Empresa;

/**
 * Capa de persistència per la gestió de l'empresa amb els seus departaments i
 * empleats, segons el model definit en el projecte Empresa.
 *
 * L'aplicació que usi la capa de persistència ha de disposar d'un fitxer de
 * propietats (veure documentació dels constructors) accessible via load(Reader)
 * amb les propietats següents:<BR>
 * className: Nom de la classe XQDataSource del driver XQJ per connectar<BR>
 * path: Ruta del fitxer XML en el que es basa la capa de persistència
 * incorporant el nom de la BD en format adequat segons el SGBD<BR>
 * transactional: Per saber si el SGBD és o no és transaccional (Y/N)<BR>
 * updateVersion: Per saber el tipus de llenguatge d'actualització a usar
 * (PL/XQUF)<BR>
 * i les propietats específiques de connexió al SGBD, segons driver a usar<BR>
 *
 * @author Isidre Guixà
 */
public class EPXQJ {

    private XQConnection con;
    private String path;

    /* Variable per control de transaccions */
    private boolean transOn;

    private String transactional;
    private String updateVersion;

    /* Variable per formatar les dates segons format requerit */
    private static SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");

    /* Variables per conèixer els objectes que tenim en memòria */
    private Empresa empresa;
    private HashMap<Integer, Empleat> hmEmp = new HashMap();
    private HashMap<Integer, Departament> hmDept = new HashMap();

    /* XQPreparedExpression */
    private XQItemType xqitStr = null;
    private XQPreparedExpression xqpeDept;

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats de nom EPXQJ.properties.
     *
     * @throws EPXQJException si hi ha algun problema en el fitxer de propietats
     * o en establir la connexió
     */
    public EPXQJ() {
        this("EPXQJ.properties");
    }

    /**
     * Constructor que estableix connexió amb el servidor a partir de les dades
     * informades en fitxer de propietats, i en cas de ser null cercarà el
     * fitxer de nom EPXQJ.properties.
     *
     * @param nomFitxerPropietats
     * @throws EPXQJException si hi ha algun problema en el fitxer de propietats
     * o en establir la connexió
     */
    public EPXQJ(String nomFitxerPropietats) {
        if (nomFitxerPropietats == null) {
            nomFitxerPropietats = "EPXQJ.properties";
        }
        Properties props = new Properties();
        try {
            props.load(new FileReader(nomFitxerPropietats));
        } catch (FileNotFoundException ex) {
            throw new EPXQJException("No es troba fitxer de propietats", ex);
        } catch (IOException ex) {
            throw new EPXQJException("Error en carregar fitxer de propietats", ex);
        }

        String className = props.getProperty("className");
        if (className == null || className.length() == 0) {
            throw new EPXQJException("Error en carregar fitxer de propietats");
        }
        props.remove("className");

        path = props.getProperty("path");
        if (path == null || path.length() == 0) {
            throw new EPXQJException("Error en carregar fitxer de propietats");
        }
        props.remove("path");

        updateVersion = props.getProperty("updateVersion");
        if (updateVersion == null) {
            throw new EPXQJException("Error en carregar fitxer de propietats");
        }
        if (!updateVersion.equals("PL") && !updateVersion.equals("XQUF")) {
            throw new EPXQJException("Error en carregar fitxer de propietats");
        }
        props.remove("updateVersion");

        transactional = props.getProperty("transactional");
        if (transactional == null) {
            throw new EPXQJException("Error en carregar fitxer de propietats");
        }
        if (!transactional.equals("N") && !transactional.equals("Y")) {
            throw new EPXQJException("Error en carregar fitxer de propietats");
        }
        props.remove("transactional");

        XQDataSource xqs;
        try {
            xqs = (XQDataSource) Class.forName(className).newInstance();
            xqs.setProperties(props);
            con = xqs.getConnection();
            if (transactional.equals("Y")) {
                // Desactivem autocommit, que per defecte, en XQJ està activat
                con.setAutoCommit(false);
            }
            System.out.println("Connexió establerta");
        } catch (XQException ex) {
            throw new EPXQJException("Error en intentar establir connexió.", ex);
        } catch (ClassNotFoundException | InstantiationException | IllegalAccessException ex) {
            throw new EPXQJException("Error en obrir la connexió.", ex);
        }

        XQExpression q = null;
        try {
            // Comprovació que el document indicat a path existeix
//              q = con.query(path); // Aquesta consulta serviria però és molt ineficient, doncs
            // efectua consulta sobre tot l'arxiu...
            q = con.createExpression();
            String cad = path + "/empresa/nom/string()";
            XQResultSequence xqrs = q.executeQuery(cad);
            // En BaseX i Sedna, una instrucció anterior "peta" si el path és erroni
            // Però en eXist-db no "peta". Simplement com que no troba el que se li demana
            // retorna un XQResultSequence "buit". Per això cal fer una consulta demanant
            // una dada que tinguem la seguretat que existeixi, com el nom de l'empresa
            // i posteriorment, via xqrs.next() veure que existeix.
            // En el cas d'eXist-db l'execute "no peta" i continua per la instruccií "if"
            // següent. Si entra dins aquest "if" podem concloure que el path és erroni
            // i provocar la mateixa excepció que en BaseX i Sedna té lloc en l'executeQuery.
            if (!xqrs.next()) {
                throw new EPXQJException("Error en comprovar l'exixtencia del path " + path);
            }
        } catch (XQException ex) {
            con = null;
            throw new EPXQJException("Error en comprovar existencia de " + path + " en el SGBD.", ex);
        } finally {
            tancarExpression(q);
        }

    }

    /**
     * Tanca la capa de persistència, tancant la connexió amb la BD.
     *
     * @throws EPXQJException si hi ha algun problema en tancar la connexió
     */
    public void closeCapa() {
        if (con != null) {
            try {
                con.close();
            } catch (XQException ex) {
                if (!ex.getMessage().contains("SE4611")) {
                    throw new EPXQJException("Error en tancar la connexió", ex);
                }
            } finally {
                con = null;
                transOn = false;
            }
        }
    }

    /**
     * Tanca la transacció activa validant els canvis a la BD.
     *
     * @throws EPXQJException si hi ha algun problema
     */
    public void commit() {
        if (transactional.equals("Y") && transOn) {
            try {
                con.commit();
                transOn = false;
            } catch (XQException ex) {
                transOn = false;
                throw new EPXQJException("Error en fer commit.", ex);
            }
        }
    }

    /**
     * Tanca la transacció activa sense validar els canvis a la BD.
     *
     * @throws EPXQJException si hi ha algun problema
     */
    public void rollback() {
        if (transactional.equals("Y") && transOn) {
            try {
                con.rollback();
                transOn = false;
            } catch (XQException ex) {
                transOn = false;
                throw new EPXQJException("Error en fer rollback.", ex);
            }
        }
    }

    /**
     * Recupera l'objecte Empresa
     *
     * @return L'objecte Empresa
     * @throws EPXQJException si hi ha algun problema en recuperar l'empresa
     */
    public Empresa getEmpresa() {
        if (empresa != null) {
            return empresa;
        }
        // Si encara no teniem l'empresa en memòria:
        XQExpression q = null;
        String nom = null;
        Calendar dataCreacio = null;
        try {
            transOn = true;
            q = con.createExpression();
            XQResultSequence xqrs = q.executeQuery(path + "/empresa/nom/string()");
            xqrs.next();
            nom = xqrs.getItemAsString(null);
            xqrs = q.executeQuery(path + "/empresa/dataCreacio/string()");
            xqrs.next();
            String aux = xqrs.getItemAsString(null);
            dataCreacio = new GregorianCalendar();
            dataCreacio.setTime(sdf.parse(aux));
        } catch (XQException ex) {
            transOn = false;
            throw new EPXQJException("Error en getEmpresa", ex);
        } catch (Exception ex) {
            throw new EPXQJException("Error en getEmpresa", ex);
        } finally {
            tancarExpression(q);
        }
        empresa = new Empresa(nom, dataCreacio);
        return empresa;
    }

    /**
     * Intenta recuperar el departament amb codi indicat per paràmetre.
     *
     * @param codi del departament a recuperar
     * @return L'objecte Departament trobat o null si no existeix
     * @throws EPXQJException si es produeix algun error (introduir un codi
     * inadequat genera també aquesta excepció)
     */
    public Departament getDepartament(int codi) {
        if (codi <= 0) {
            throw new EPXQJException("Intent de recuperar departament amb codi " + codi + " erroni");
        }
        Departament d = hmDept.get(codi);
        if (d != null) {
            return d;
        }
        // Si encara no teniem el departament en memòria:
        if (xqpeDept == null) {
            String cad = "declare variable $codi external;";
            cad = cad + path + "//departaments/dept[@codi=$codi]";
            try {
                xqpeDept = con.prepareExpression(cad);
            } catch (XQException ex) {
                transOn = false;
                throw new EPXQJException("Error en crear el prepared expression xqpeDept");
            }
        }

        String aux = "d" + codi;
        try {
            transOn = true;
            xqpeDept.bindString(new QName("codi"), aux,xqitStr);
            XQResultSequence xqrs = xqpeDept.executeQuery();
            if (xqrs.next()==false) {
                return null;
            }
            String resultat = xqrs.getItemAsString(null);
            // Dins "resultat" hi ha un node <dept>
            SAXBuilder sb = new SAXBuilder();
            Document doc = sb.build(new StringReader(resultat));
            Element xDept = doc.getRootElement();
            String nom = xDept.getChildText("nom");
            String loc = xDept.getChildText("localitat");
            d = new Departament(codi, nom, loc);
            hmDept.put(codi, d);
            return d;
        } catch (XQException ex) {
            transOn = false;
            throw new EPXQJException("Error en intentar recuperar departament de codi " + codi, ex);
        } catch (Exception ex) {
            throw new EPXQJException("Error en intentar recuperar departament de codi " + codi, ex);
        } 
    }

    /**
     * Intenta recuperar l'empleat amb codi indicat per paràmetre.
     *
     * @param codi de l'empleat a recuperar
     * @return L'objecte Empleat trobat o null si no existeix
     * @throws EPXQJException si es produeix algun error (introduir un codi
     * inadequat genera també aquesta excepció)
     */
    public Empleat getEmpleat(int codi) {
        /* TO DO */
        return null;
    }

    /**
     * Informa de l'existència d'empleat amb el codi indicat per paràmetre.
     *
     * @return Valor booleà que indica l'existència de l'empleat indicat
     * @throws EPXQJException si es produeix algun error (introduir un codi
     * inadequat genera també aquesta excepció)
     */
    public boolean existeixEmpleat(int codi) {
        /* TO DO */
        return false;
    }

    /**
     * Facilita el nombre de subordinats d'un empleat
     *
     * @param codi de l'empleat al què comptabilitzar els subordinats
     * @return El nombre de subordinats de l'empleat
     * @throws EPXQJException si es produeix algun error (introduir un codi
     * inadequat o la inexistència d'empleat amb el codi indicat genera també
     * aquesta excepció)
     */
    public int comptarSubordinats(int codi) {
        /* TO DO */
        return 0;
    }

    /**
     * Comprova si un empleat és subordinat directe o indirecte (varis nivells)
     * del cap
     *
     * @param cap
     * @param emp
     * @return true si emp és subordinat (directe o indirecte) de cap i false en
     * cas contrari
     */
    public boolean esSubordinatDirecteIndirecte(int cap, int emp) {
        /* TO DO */
        return false;
    }

    /**
     * Intenta eliminar un empleat de codi indicat per paràmetre.
     *
     * @param codi de l'empleat a eliminar
     * @param actCap: Indica l'actuació a tenir en compte vers els empleats dels
     * que pugui ser el cap. No té cap efecte si l'empleat no és cap de ningú.
     * Actuacions:<BR>
     * Valor 0: Deixar els empleats que el tenen per cap, "SENSE" cap<BR>
     * Valor &lt;0: Generar excepció indicant que no ha estat possible
     * l'eliminació per ser cap d'altres empleats<BR>
     * Valor &gt;0: Canviar el cap dels empleats afectats per l'empleat amb codi
     * = al valor del paràmetre actCap (que ha d'existir)<BR>
     *
     */
    public void eliminarEmpleat(int codi, int actCap) {
        /* TO DO */
    }

    /**
     *
     * MÈTODES PRIVATS, d'ajut pel desenvolupament de la capa
     *
     */
    private void tancarExpression(XQExpression xqe) {
        try {
            if (xqe != null) {
                xqe.close();
            }
        } catch (XQException ex) {
            throw new EPXQJException("Error al tancar l'expressió", ex);
        }
    }
}
